# Dictionaries used to map column names to valid argument names

LEAD_COLUMN_MAP = {
    "first_name": ["first", "fn"],
    "last_name": ["last", "ln", "lastname"],
    "phone_number": ["phone", "cell", "cellphone"],
    "email": ["emailaddress"],
    "follow_up": ["follow"],
}
