folder_json = {
    "folders": [
        {
            "id": "123",
            "name": "Voter Events",
            "parent": "0",
            "permissions": "full",
            "subfolders": [
                {
                    "id": "200",
                    "name": "Phone Banks",
                    "parent": "123",
                    "permissions": "full",
                },
                {
                    "id": "201",
                    "name": "Text Banks",
                    "parent": "121",
                    "permissions": "full",
                },
            ],
        },
        {"id": "121", "name": "Fundraising", "parent": "0", "permissions": "full"},
        {"id": "122", "name": "Internal", "parent": "0", "permissions": "full"},
    ],
    "total": 3,
}
form_json = {
    "forms": [
        {
            "id": "1001",
            "created": "2018-08-01 10:10:00",
            "db": "1",
            "deleted": "0",
            "folder": "123",
            "language": "en",
            "name": "Test Form 1",
            "num_columns": "3",
            "progress_meter": "0",
            "submissions": "1321",
            "submissions_unread": "1",
            "updated": "2019-10-10 11:04:53",
            "viewkey": "viewkey_test",
            "views": "1000",
            "submissions_today": 0,
            "last_submission_id": "54321",
            "last_submission_time": "2020-10-10 10:10:10",
            "url": "https://test.formstack.com/forms/test_1",
            "encrypted": False,
            "thumbnail_url": None,
            "submit_button_title": "Submit",
            "inactive": True,
            "timezone": "US/Eastern",
            "should_display_one_question_at_a_time": False,
            "can_access_1q_feature": True,
            "is_workflow_form": False,
            "is_workflow_published": False,
            "has_approvers": False,
            "edit_url": "https://www.formstack.com/admin/form/builder/1001/build",
            "data_url": "",
            "summary_url": "",
            "rss_url": "",
            "permissions": 150,
            "can_edit": True,
        },
        {
            "id": "1002",
            "created": "2018-08-01 10:10:00",
            "db": "1",
            "deleted": "0",
            "folder": "123",
            "language": "en",
            "name": "Test Form 2",
            "num_columns": "3",
            "progress_meter": "0",
            "submissions": "1321",
            "submissions_unread": "1",
            "updated": "2019-10-10 11:04:53",
            "viewkey": "viewkey_test",
            "views": "1000",
            "submissions_today": 0,
            "last_submission_id": "54321",
            "last_submission_time": "2020-10-10 10:10:10",
            "url": "https://test.formstack.com/forms/test_2",
            "encrypted": False,
            "thumbnail_url": None,
            "submit_button_title": "Submit",
            "inactive": True,
            "timezone": "US/Eastern",
            "should_display_one_question_at_a_time": False,
            "can_access_1q_feature": True,
            "is_workflow_form": False,
            "is_workflow_published": False,
            "has_approvers": False,
            "edit_url": "https://www.formstack.com/admin/form/builder/1002/build",
            "data_url": "",
            "summary_url": "",
            "rss_url": "",
            "permissions": 150,
            "can_edit": True,
        },
    ],
    "total": 3,
}
submission_id = 332525567
submission_json = {
    "id": str(submission_id),
    "timestamp": "2017-06-13 17:07:51",
    "user_agent": "Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko",
    "remote_addr": "50.232.88.142",
    "payment_status": "",
    "form": "2691007",
    "latitude": "39.955799102783",
    "longitude": "-105.16829681396",
    "data": [
        {"field": "52404588", "value": "95386.00000"},
        {"field": "52404587", "value": "lisetalbott@gmail.com"},
    ],
    "pretty_field_id": "52404587",
}
form_submissions_json = {
    "submissions": [
        {
            "id": "325343738",
            "timestamp": "2017-05-07 21:52:42",
            "user_agent": (
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_4) AppleWebKit/537.36"
                " (KHTML, like Gecko) Chrome/57.0.2987.133 Safari/537.36"
            ),
            "remote_addr": "73.163.190.175",
            "payment_status": "",
            "latitude": "38.905700683594",
            "longitude": "-76.978302001953",
            "read": "1",
        },
        {
            "id": "325512020",
            "timestamp": "2017-05-08 15:31:08",
            "user_agent": (
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/601.7.8"
                " (KHTML, like Gecko) Version/9.1.3 Safari/601.7.8"
            ),
            "remote_addr": "65.242.21.250",
            "payment_status": "",
            "latitude": "37.750999450684",
            "longitude": "-97.821998596191",
            "read": "1",
        },
        {
            "id": "325512197",
            "timestamp": "2017-05-08 15:31:52",
            "user_agent": (
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_4) AppleWebKit/603.1.30"
                " (KHTML, like Gecko) Version/10.1 Safari/603.1.30"
            ),
            "remote_addr": "65.242.21.250",
            "payment_status": "",
            "latitude": "37.750999450684",
            "longitude": "-97.821998596191",
            "read": "1",
        },
    ],
    "total": 1321,
    "pages": 53,
    "pretty_field_id": "52404587",
}
form_fields_json = [
    {
        "id": "52404588",
        "label": "Zip",
        "hide_label": "0",
        "description": "",
        "name": "zip",
        "type": "number",
        "options": "",
        "required": "1",
        "uniq": "0",
        "hidden": "0",
        "readonly": "0",
        "colspan": "1",
        "sort": "0",
        "logic": None,
        "calculation": "",
        "workflow_access": "write",
        "default": "",
        "field_one_calculation": 0,
        "field_two_calculation": 0,
        "calculation_units": "",
        "calculation_operator": "",
        "calculation_type": "",
        "calculation_category": "",
        "calculation_allow_negatives": 0,
        "text_size": 5,
        "min_value": "",
        "max_value": "",
        "currency": "",
        "decimals": "5",
        "slider": "0",
        "placeholder": "",
    },
    {
        "id": "52404587",
        "label": "Email",
        "hide_label": "0",
        "description": "",
        "name": "email",
        "type": "email",
        "options": "",
        "required": "1",
        "uniq": "0",
        "hidden": "0",
        "readonly": "0",
        "colspan": "1",
        "sort": "1",
        "logic": None,
        "calculation": "",
        "workflow_access": "write",
        "default": "",
        "text_size": 50,
        "maxlength": 0,
        "confirm": 0,
        "confirmationText": "",
        "placeholder": "",
    },
]
